# -*- coding:utf-8 -*-
"""
Implements a function for hoisting specially constructed comments in SQL
queries and using them to rewrite that query. This is done to support query
hints whilst obviating patching Django's ORM in complex ways.
"""
from __future__ import unicode_literals

import operator
import re
from collections import OrderedDict

from django.utils import six

# The rewrite comments contain a single quote mark that would need be escaped
# if entered in a column name or something like that. We aren't too worried
# about SQL injection since data is not in given the statement - it's still in
# params
REWRITE_MARKER = "/*QueryRewrite':"

# Regex to match a rewrite rule
query_rewrite_re = re.compile(r"/\*QueryRewrite':(.*?)\*/")
# Regex to parse an index hint into a tuple
index_rule_re = re.compile(
    r"""
    index=
    (?P<table_name>`[^`]+`)
    \ # space
    (?P<rule>USE|IGNORE|FORCE)
    \ # space
    (
        FOR
        \ # space
        (?P<for_what>JOIN|ORDER\ BY|GROUP\ BY)
        \ # space
    )?
    (?P<index_names>(`[^`]+`(,`[^`]+`)*)|NONE)
    """,
    re.VERBOSE
)


def rewrite_query(sql):
    comments = []
    hints = []
    index_hints = []
    for match in query_rewrite_re.findall(sql):
        if match in SELECT_HINT_TOKENS:
            hints.append(match)
        elif match.startswith('label='):
            comments.append(match[6:])
        elif match.startswith('index='):
            # Extra parsing
            index_match = index_rule_re.match(match)
            if index_match:
                index_hints.append((
                    index_match.group('table_name'),
                    index_match.group('rule'),
                    index_match.group('index_names'),
                    index_match.group('for_what'),
                ))

        # Silently fail on unrecognized rewrite requests

    # Delete all rewrite comments
    sql = query_rewrite_re.sub('', sql)

    if comments or hints or index_hints:  # If nothing to do, don't bother
        sql = modify_sql(sql, comments, hints, index_hints)

    return sql


# A translation of the grammar for SELECT - all the possible hints that can
# appear afterwards
SELECT_HINTS = OrderedDict([
    ('distinctness', ('ALL', 'DISTINCT', 'DISTINCTROW',)),
    ('priority', ('HIGH_PRIORITY',)),
    ('join_order', ('STRAIGHT_JOIN',)),
    ('result_size', ('SQL_SMALL_RESULT', 'SQL_BIG_RESULT',)),
    ('buffer_result', ('SQL_BUFFER_RESULT',)),
    ('query_cache', ('SQL_CACHE', 'SQL_NO_CACHE',)),
    ('found_rows', ('SQL_CALC_FOUND_ROWS',)),
])

# Any pre-expression tokens that are query hints
SELECT_HINT_TOKENS = frozenset(
    six.moves.reduce(operator.add, SELECT_HINTS.values())
)

# Don't go crazy reading this - it's just templating a piece of the below regex
hints_re_piece = '\n'.join(
    r'(?P<{group_name}>({tokens})\s+)?'.format(
        group_name=group_name,
        tokens='|'.join(token_set)
    )
    for group_name, token_set in six.iteritems(SELECT_HINTS)
)


# This is the one big regex that parses the start of the SQL statement
# It makes a few assumptions that are valid for queries from the Django ORM but
# may not be for other queries - for example, comments are only searched for
# immediately after the first token, pretty much assuming they could only have
# come from add_sql_piece
query_start_re = re.compile(
    r"""
        ^
        \s*
        (?P<keyword>SELECT|UPDATE|DELETE)
        # comments - N times /*a*/whitespace
        (?P<comments>(\s*/\*.*?\*/\s*)+|\s+)
    """ + hints_re_piece,
    re.VERBOSE | re.IGNORECASE
)


def modify_sql(sql, add_comments, add_hints, add_index_hints):
    """
    Parse the start of the SQL, injecting each string in add_comments in
    individual SQL comments after the first keyword, and adding the named
    SELECT hints from add_hints, taking the latest in the list in cases of
    multiple mutually exclusive hints being given
    """
    match = query_start_re.match(sql)
    if not match:
        # We don't understand what kind of query this is, don't rewrite it
        return sql

    tokens = [match.group('keyword')]
    comments = match.group('comments').strip()
    if comments:
        tokens.append(comments)

    # Inject comments after all existing comments
    for comment in add_comments:
        tokens.append('/*{}*/'.format(comment))

    # Don't bother with SELECT rewrite rules on non-SELECT queries
    if tokens[0] == "SELECT":

        for group_name, hint_set in six.iteritems(SELECT_HINTS):

            try:
                # Take the last hint we were told to add from this hint_set
                to_add = [hint for hint in add_hints if hint in hint_set][-1]
                tokens.append(to_add)
            except IndexError:
                # We weren't told to add any, so just add any hint from this
                # set that was already there
                existing = match.group(group_name)
                if existing is not None:
                    tokens.append(existing.rstrip())

    # Maybe rewrite the remainder of the statement for index hints
    remainder = sql[match.end():]

    if tokens[0] == "SELECT" and add_index_hints:
        for index_hint in add_index_hints:
            remainder = modify_sql_index_hints(remainder, *index_hint)

    # Join everything
    tokens.append(remainder)
    return ' '.join(tokens)


table_spec_re_template = r'''
    \b(?P<operator>FROM|JOIN)
    \s+
    {table_name}
    \s+
'''

replacement_template = (
    r'\g<operator> {table_name} '
    r'{rule} INDEX {for_section}({index_names}) '
)


def modify_sql_index_hints(sql, table_name, rule, index_names, for_what):
    table_spec_re = table_spec_re_template.format(table_name=table_name)
    if for_what:
        for_section = 'FOR {} '.format(for_what)
    else:
        for_section = ''
    replacement = replacement_template.format(
        table_name=table_name,
        rule=rule,
        for_section=for_section,
        index_names=('' if index_names == 'NONE' else index_names)
    )
    return re.sub(table_spec_re, replacement, sql, count=1, flags=re.VERBOSE)



# -*- coding:utf-8 -*-
from __future__ import absolute_import

from django.core import checks
from django.db.models import CharField, IntegerField, Lookup, TextField
from django.utils import six
from django.utils.translation import ugettext_lazy as _

from django_mysql.forms import SimpleListField
from django_mysql.models.lookups import SetContains, SetIContains
from django_mysql.models.transforms import SetLength
from django_mysql.validators import ListMaxLengthValidator


class ListFieldMixin(object):

    def __init__(self, base_field, size=None, **kwargs):
        self.base_field = base_field
        self.size = size

        super(ListFieldMixin, self).__init__(**kwargs)

        if self.size:
            self.validators.append(ListMaxLengthValidator(int(self.size)))

    def get_default(self):
        default = super(ListFieldMixin, self).get_default()
        if default == '':
            return []
        else:
            return default

    def check(self, **kwargs):
        errors = super(ListFieldMixin, self).check(**kwargs)
        if not isinstance(self.base_field, (CharField, IntegerField)):
            errors.append(
                checks.Error(
                    'Base field for list must be a CharField or IntegerField.',
                    hint=None,
                    obj=self,
                    id='django_mysql.E005'
                )
            )
            return errors

        # Remove the field name checks as they are not needed here.
        base_errors = self.base_field.check()
        if base_errors:
            messages = '\n    '.join(
                '%s (%s)' % (error.msg, error.id)
                for error in base_errors
            )
            errors.append(
                checks.Error(
                    'Base field for list has errors:\n    %s' % messages,
                    hint=None,
                    obj=self,
                    id='django_mysql.E004'
                )
            )
        return errors

    @property
    def description(self):
        return _('List of %(base_description)s') % {
            'base_description': self.base_field.description
        }

    def set_attributes_from_name(self, name):
        super(ListFieldMixin, self).set_attributes_from_name(name)
        self.base_field.set_attributes_from_name(name)

    def deconstruct(self):
        name, path, args, kwargs = super(ListFieldMixin, self).deconstruct()

        bad_paths = (
            'django_mysql.models.fields.lists.' + self.__class__.__name__,
            'django_mysql.models.fields.' + self.__class__.__name__
        )
        if path in bad_paths:
            path = 'django_mysql.models.' + self.__class__.__name__

        args.insert(0, self.base_field)
        kwargs['size'] = self.size
        return name, path, args, kwargs

    def to_python(self, value):
        if isinstance(value, six.string_types):
            if not len(value):
                value = []
            else:
                value = [self.base_field.to_python(v) for
                         v in value.split(',')]
        return value

    def from_db_value(self, value, expression, connection, context):
        # Similar to to_python, for Django 1.8+
        if isinstance(value, six.string_types):
            if not len(value):
                value = []
            else:
                value = [self.base_field.to_python(v) for
                         v in value.split(',')]
        return value

    def get_prep_value(self, value):
        if isinstance(value, list):
            value = [
                six.text_type(self.base_field.get_prep_value(v))
                for v in value
            ]
            for v in value:
                if ',' in v:
                    raise ValueError(
                        "List members in {klass} {name} cannot contain commas"
                        .format(klass=self.__class__.__name__,
                                name=self.name)
                    )
                elif not len(v):
                    raise ValueError(
                        "The empty string cannot be stored in {klass} {name}"
                        .format(klass=self.__class__.__name__,
                                name=self.name)
                    )
            return ','.join(value)
        return value

    def get_lookup(self, lookup_name):
        lookup = super(ListFieldMixin, self).get_lookup(lookup_name)
        if lookup:
            return lookup

        try:
            index = int(lookup_name)
        except ValueError:
            pass
        else:
            index += 1  # MySQL uses 1-indexing
            return IndexLookupFactory(index)

        return lookup

    def value_to_string(self, obj):
        vals = self._get_val_from_obj(obj)
        return self.get_prep_value(vals)

    def formfield(self, **kwargs):
        defaults = {
            'form_class': SimpleListField,
            'base_field': self.base_field.formfield(),
            'max_length': self.size,
        }
        defaults.update(kwargs)
        return super(ListFieldMixin, self).formfield(**defaults)

    def contribute_to_class(self, cls, name, **kwargs):
        super(ListFieldMixin, self).contribute_to_class(cls, name, **kwargs)
        self.base_field.model = cls


class ListCharField(ListFieldMixin, CharField):
    """
    A subclass of CharField for using MySQL's handy FIND_IN_SET function with.
    """
    def check(self, **kwargs):
        errors = super(ListCharField, self).check(**kwargs)

        # Unfortunately this check can't really be done for IntegerFields since
        # they have boundless length
        has_base_error = any(e.id == 'django_mysql.E004' for e in errors)
        if (
            not has_base_error and
            isinstance(self.base_field, CharField) and
            self.size
        ):
            max_size = (
                # The chars used
                (self.size * (self.base_field.max_length)) +
                # The commas
                self.size - 1
            )
            if max_size > self.max_length:
                errors.append(
                    checks.Error(
                        'Field can overrun - set contains CharFields of max '
                        'length %s, leading to a comma-combined max length of '
                        '%s, which is greater than the space reserved for the '
                        'set - %s' %
                        (self.base_field.max_length, max_size,
                            self.max_length),
                        hint=None,
                        obj=self,
                        id='django_mysql.E006'
                    )
                )
        return errors


class ListTextField(ListFieldMixin, TextField):
    pass


ListCharField.register_lookup(SetContains)
ListTextField.register_lookup(SetContains)

ListCharField.register_lookup(SetIContains)
ListTextField.register_lookup(SetIContains)

ListCharField.register_lookup(SetLength)
ListTextField.register_lookup(SetLength)


class IndexLookup(Lookup):

    def __init__(self, index, *args, **kwargs):
        super(IndexLookup, self).__init__(*args, **kwargs)
        self.index = index

    def as_sql(self, qn, connection):
        lhs, lhs_params = self.process_lhs(qn, connection)
        rhs, rhs_params = self.process_rhs(qn, connection)
        params = lhs_params + rhs_params
        # Put rhs on the left since that's the order FIND_IN_SET uses
        return '(FIND_IN_SET(%s, %s) = %s)' % (rhs, lhs, self.index), params


class IndexLookupFactory(object):

    def __init__(self, index):
        self.index = index

    def __call__(self, *args, **kwargs):
        return IndexLookup(self.index, *args, **kwargs)



# -*- coding:utf-8 -*-
from __future__ import unicode_literals

from django.db.models import F, Value
from django.db.models.expressions import BaseExpression

from django_mysql.utils import collapse_spaces


class TwoSidedExpression(BaseExpression):

    def __init__(self, lhs, rhs):
        super(TwoSidedExpression, self).__init__()
        self.lhs = lhs
        self.rhs = rhs

    def get_source_expressions(self):
        return [self.lhs, self.rhs]

    def set_source_expressions(self, exprs):
        self.lhs, self.rhs = exprs


class ListF(object):
    def __init__(self, field_name):
        self.field_name = field_name
        self.field = F(field_name)

    def append(self, value):
        if not hasattr(value, 'as_sql'):
            value = Value(value)
        return AppendListF(self.field, value)

    def appendleft(self, value):
        if not hasattr(value, 'as_sql'):
            value = Value(value)
        return AppendLeftListF(self.field, value)

    def pop(self):
        return PopListF(self.field)

    def popleft(self):
        return PopLeftListF(self.field)


class AppendListF(TwoSidedExpression):

    # A slightly complicated expression.
    # basically if 'value' is not in the set, concat the current set with a
    # comma and 'value'
    # N.B. using MySQL side variables to avoid repeat calculation of
    # expression[s]
    sql_expression = collapse_spaces("""
        CONCAT_WS(
            ',',
            IF(
                (@tmp_f:=%s) > '',
                @tmp_f,
                NULL
            ),
            %s
        )
    """)

    def as_sql(self, compiler, connection):
        field, field_params = compiler.compile(self.lhs)
        value, value_params = compiler.compile(self.rhs)

        sql = self.sql_expression % (field, value)

        params = []
        params.extend(value_params)
        params.extend(field_params)

        return sql, params


class AppendLeftListF(TwoSidedExpression):

    # A slightly complicated expression.
    # basically if 'value' is not in the set, concat the current set with a
    # comma and 'value'
    # N.B. using MySQL side variables to avoid repeat calculation of
    # expression[s]
    sql_expression = collapse_spaces("""
        CONCAT_WS(
            ',',
            %s,
            IF(
                (@tmp_f:=%s) > '',
                @tmp_f,
                NULL
            )
        )
    """)

    def as_sql(self, compiler, connection):
        field, field_params = compiler.compile(self.lhs)
        value, value_params = compiler.compile(self.rhs)

        sql = self.sql_expression % (value, field)

        params = []
        params.extend(field_params)
        params.extend(value_params)

        return sql, params


class PopListF(BaseExpression):

    sql_expression = collapse_spaces("""
        SUBSTRING(
            @tmp_f:=%s,
            1,
            IF(
                LOCATE(',', @tmp_f),
                (
                    CHAR_LENGTH(@tmp_f) -
                    CHAR_LENGTH(SUBSTRING_INDEX(@tmp_f, ',', -1)) -
                    1
                ),
                0
            )
        )
    """)

    def __init__(self, lhs):
        super(BaseExpression, self).__init__()
        self.lhs = lhs

    def get_source_expressions(self):
        return [self.lhs]

    def set_source_expressions(self, exprs):
        self.lhs = exprs[0]

    def as_sql(self, compiler, connection):
        field, field_params = compiler.compile(self.lhs)

        sql = self.sql_expression % (field)
        return sql, field_params


class PopLeftListF(BaseExpression):

    sql_expression = collapse_spaces("""
        IF(
            (@tmp_c:=LOCATE(',', @tmp_f:=%s)) > 0,
            SUBSTRING(@tmp_f, @tmp_c + 1),
            ''
        )
    """)

    def __init__(self, lhs):
        super(BaseExpression, self).__init__()
        self.lhs = lhs

    def get_source_expressions(self):
        return [self.lhs]

    def set_source_expressions(self, exprs):
        self.lhs = exprs[0]

    def as_sql(self, compiler, connection):
        field, field_params = compiler.compile(self.lhs)

        sql = self.sql_expression % (field)
        return sql, field_params


class SetF(object):

    def __init__(self, field_name):
        self.field = F(field_name)

    def add(self, value):
        if not hasattr(value, 'as_sql'):
            value = Value(value)
        return AddSetF(self.field, value)

    def remove(self, value):
        if not hasattr(value, 'as_sql'):
            value = Value(value)
        return RemoveSetF(self.field, value)


class AddSetF(TwoSidedExpression):

    # A slightly complicated expression.
    # basically if 'value' is not in the set, concat the current set with a
    # comma and 'value'
    # N.B. using MySQL side variables to avoid repeat calculation of
    # expression[s]
    sql_expression = collapse_spaces("""
        IF(
            FIND_IN_SET(@tmp_val:=%s, @tmp_f:=%s),
            @tmp_f,
            CONCAT_WS(
                ',',
                IF(CHAR_LENGTH(@tmp_f), @tmp_f, NULL),
                @tmp_val
            )
        )
    """)

    def as_sql(self, compiler, connection):
        field, field_params = compiler.compile(self.lhs)
        value, value_params = compiler.compile(self.rhs)

        sql = self.sql_expression % (value, field)

        params = []
        params.extend(value_params)
        params.extend(field_params)

        return sql, params


class RemoveSetF(TwoSidedExpression):

    # Wow, this is a real doozy of an expression.
    # Basically, if it IS in the set, cut the string up to be everything except
    # that element.
    # There are some tricks going on - e.g. LEAST to evaluate a sub expression
    # but not use it in the output of CONCAT_WS
    sql_expression = collapse_spaces("""
        IF(
            @tmp_pos:=FIND_IN_SET(%s, @tmp_f:=%s),
            CONCAT_WS(
                ',',
                LEAST(
                    @tmp_len:=(
                        CHAR_LENGTH(@tmp_f) -
                        CHAR_LENGTH(REPLACE(@tmp_f, ',', '')) +
                        IF(CHAR_LENGTH(@tmp_f), 1, 0)
                    ),
                    NULL
                ),
                CASE WHEN
                    (@tmp_before:=SUBSTRING_INDEX(@tmp_f, ',', @tmp_pos - 1))
                    = ''
                    THEN NULL
                    ELSE @tmp_before
                END,
                CASE WHEN
                    (@tmp_after:=
                        SUBSTRING_INDEX(@tmp_f, ',', - (@tmp_len - @tmp_pos)))
                    = ''
                    THEN NULL
                    ELSE @tmp_after
                END
            ),
            @tmp_f
        )
    """)

    def as_sql(self, compiler, connection):
        field, field_params = compiler.compile(self.lhs)
        value, value_params = compiler.compile(self.rhs)

        sql = self.sql_expression % (value, field)

        params = []
        params.extend(value_params)
        params.extend(field_params)

        return sql, params



# -*- coding:utf-8 -*-
from __future__ import unicode_literals

import mock
from unittest import SkipTest

import pytest
from django.db import connection, connections
from django.db.models import F
from django.test import TestCase

from django_mysql import forms
from django_mysql.models import JSONField
from testapp.models import JSONModel, TemporaryModel
from testapp.utils import print_all_queries


class JSONFieldTestCase(TestCase):

    @classmethod
    def setUpClass(cls):
        if not (
            not connection.is_mariadb and
            connection.mysql_version >= (5, 7)
        ):
            raise SkipTest("JSONField requires MySQL 5.7+")
        super(JSONFieldTestCase, cls).setUpClass()


class TestSaveLoad(JSONFieldTestCase):

    def test_empty_dict(self):
        m = JSONModel()
        assert m.attrs == {}
        m.save()
        m = JSONModel.objects.get()
        assert m.attrs == {}

    def test_values(self):
        m = JSONModel(attrs={'key': 'value'})
        assert m.attrs == {'key': 'value'}
        m.save()
        m = JSONModel.objects.get()
        assert m.attrs == {'key': 'value'}

    def test_list(self):
        m = JSONModel(attrs=[1, 2, 4])
        assert m.attrs == [1, 2, 4]
        m.save()
        m = JSONModel.objects.get()
        assert m.attrs == [1, 2, 4]

    def test_null(self):
        m = JSONModel(attrs=None)
        assert m.attrs is None
        m.save()
        m = JSONModel.objects.get()
        assert m.attrs is None

    def test_control_characters(self):
        chars = ''.join(chr(i) for i in range(32))
        m = JSONModel(attrs=[chars])
        assert m.attrs == [chars]
        m.save()
        m = JSONModel.objects.get()
        assert m.attrs == [chars]


class QueryTests(JSONFieldTestCase):

    def setUp(self):
        super(QueryTests, self).setUp()
        JSONModel.objects.bulk_create([
            JSONModel(attrs={'a': 'b'}),
            JSONModel(attrs=1337),
            JSONModel(attrs=['an', 'array']),
            JSONModel(attrs=None),
        ])
        self.objs = list(JSONModel.objects.all().order_by('id'))

    def test_equal(self):
        assert (
            list(JSONModel.objects.filter(attrs={'a': 'b'})) ==
            [self.objs[0]]
        )

    def test_equal_value(self):
        assert (
            list(JSONModel.objects.filter(attrs=1337)) ==
            [self.objs[1]]
        )

    def test_equal_array(self):
        assert (
            list(JSONModel.objects.filter(attrs=['an', 'array'])) ==
            [self.objs[2]]
        )

    def test_equal_no_match(self):
        assert (
            list(JSONModel.objects.filter(attrs={'c': 'z'})) ==
            []
        )

    def test_equal_F_attrs(self):
        assert (
            list(JSONModel.objects.filter(attrs=F('attrs'))) ==
            [self.objs[0], self.objs[1], self.objs[2]]
        )

    def test_isnull_True(self):
        assert (
            list(JSONModel.objects.filter(attrs__isnull=True)) ==
            [self.objs[3]]
        )

    def test_isnull_False(self):
        assert (
            list(JSONModel.objects.filter(attrs__isnull=False)) ==
            [self.objs[0], self.objs[1], self.objs[2]]
        )

    def test_range_broken(self):
        with pytest.raises(NotImplementedError) as excinfo:
            JSONModel.objects.filter(attrs__range=[1, 2])

        assert (
            "Lookup 'range' doesn't work with JSONField" in str(excinfo.value)
        )


class ArrayQueryTests(JSONFieldTestCase):

    def setUp(self):
        super(ArrayQueryTests, self).setUp()
        JSONModel.objects.bulk_create([
            JSONModel(attrs=[1, 3]),
            JSONModel(attrs=[1, 3, 3]),
            JSONModel(attrs=[1, 3, 3, 7]),
            JSONModel(attrs=[2, 4]),
        ])
        self.objs = list(JSONModel.objects.all().order_by('id'))

    def test_lt_1(self):
        assert (
            list(JSONModel.objects.filter(attrs__lt=[1])) ==
            []
        )

    def test_lt_3(self):
        assert (
            list(JSONModel.objects.filter(attrs__lt=[3])) ==
            self.objs
        )

    def test_lte_1_3_3(self):
        assert (
            list(JSONModel.objects.filter(attrs__lte=[1, 3, 3])) ==
            [self.objs[0], self.objs[1]]
        )

    def test_lte_1(self):
        assert (
            list(JSONModel.objects.filter(attrs__lte=[1])) ==
            []
        )

    def test_gt_1(self):
        assert (
            list(JSONModel.objects.filter(attrs__gt=[1])) ==
            self.objs
        )

    def test_gt_1_3(self):
        assert (
            list(JSONModel.objects.filter(attrs__gt=[1, 3])) ==
            self.objs[1:]
        )

    def test_gt_2_5(self):
        assert (
            list(JSONModel.objects.filter(attrs__gt=[2, 5])) ==
            []
        )

    def test_gte_1_3(self):
        assert (
            list(JSONModel.objects.filter(attrs__gte=[1, 3])) ==
            self.objs
        )

    def test_gte_2(self):
        assert (
            list(JSONModel.objects.filter(attrs__gte=[2])) ==
            [self.objs[3]]
        )


class ExtraLookupsQueryTests(JSONFieldTestCase):

    def setUp(self):
        super(ExtraLookupsQueryTests, self).setUp()

        self.objs = [
            JSONModel.objects.create(attrs={}),
            JSONModel.objects.create(attrs={
                'a': 'b',
                'c': 1,
                '"': 'awkward',
                '\n': 'super awkward'
            }),
            JSONModel.objects.create(attrs={
                'a': 'b',
                'c': 1,
                'd': ['e', {'f': 'g'}],
                'h': True,
                'i': False,
                'j': None,
                'k': {'l': 'm'},
            }),
            JSONModel.objects.create(attrs=[1, [2]]),
            JSONModel.objects.create(attrs={
                'k': True,
                'l': False,
                '\\': 'awkward'
            }),
        ]

    def test_has_key_invalid_type(self):
        with pytest.raises(ValueError) as excinfo:
            JSONModel.objects.filter(attrs__has_key=1)
        assert "'has_key' lookup only works with" in str(excinfo.value)

    def test_has_key(self):
        assert (
            list(JSONModel.objects.filter(attrs__has_key='a')) ==
            [self.objs[1], self.objs[2]]
        )

    def test_has_key_2(self):
        assert (
            list(JSONModel.objects.filter(attrs__has_key='l')) ==
            [self.objs[4]]
        )

    def test_has_key_awkward(self):
        assert (
            list(JSONModel.objects.filter(attrs__has_key='"')) ==
            [self.objs[1]]
        )

    def test_has_key_awkward_2(self):
        assert (
            list(JSONModel.objects.filter(attrs__has_key='\n')) ==
            [self.objs[1]]
        )

    def test_has_key_awkward_3(self):
        assert (
            list(JSONModel.objects.filter(attrs__has_key='\\')) ==
            [self.objs[4]]
        )

    def test_has_keys_invalid_type(self):
        with pytest.raises(ValueError) as excinfo:
            JSONModel.objects.filter(attrs__has_keys={'a': 'b'})
        assert "only works with Sequences" in str(excinfo.value)

    def test_has_keys(self):
        with print_all_queries():
            assert (
                list(JSONModel.objects.filter(attrs__has_keys=['a', 'c'])) ==
                [self.objs[1], self.objs[2]]
            )

    def test_has_keys_2(self):
        assert (
            list(JSONModel.objects.filter(attrs__has_keys=['l'])) ==
            [self.objs[4]]
        )

    def test_has_keys_awkward(self):
        assert (
            list(JSONModel.objects.filter(attrs__has_keys=['\n', '"'])) ==
            [self.objs[1]]
        )

    def test_has_any_keys_invalid_type(self):
        with pytest.raises(ValueError) as excinfo:
            JSONModel.objects.filter(attrs__has_any_keys={'a': 'b'})
        assert "only works with Sequences" in str(excinfo.value)

    def test_has_any_keys(self):
        assert (
            list(JSONModel.objects.filter(attrs__has_any_keys=['a', 'k'])) ==
            [self.objs[1], self.objs[2], self.objs[4]]
        )

    def test_has_any_keys_awkward(self):
        assert (
            list(JSONModel.objects.filter(attrs__has_any_keys=['\\'])) ==
            [self.objs[4]]
        )

    def test_contains(self):
        assert (
            list(JSONModel.objects.filter(attrs__contains={'a': 'b'})) ==
            [self.objs[1], self.objs[2]]
        )

    def test_contains_2(self):
        assert (
            list(JSONModel.objects.filter(attrs__contains={'l': False})) ==
            [self.objs[4]]
        )

    def test_contains_array(self):
        assert (
            list(JSONModel.objects.filter(attrs__contains=[[2]])) ==
            [self.objs[3]]
        )

    def test_contained_by(self):
        assert (
            list(JSONModel.objects.filter(attrs__contained_by={
                'k': True,
                'l': False,
                '\\': 'awkward',
                'spare_key': ['unused', 'array'],
            })) ==
            [self.objs[0], self.objs[4]]
        )

    def test_contained_by_array(self):
        assert (
            list(JSONModel.objects.filter(attrs__contained_by=[1, [2], 3])) ==
            [self.objs[3]]
        )

    def test_length_lookup(self):
        assert (
            list(JSONModel.objects.filter(attrs__length=0)) ==
            [self.objs[0]]
        )

    def test_length_lookup_2(self):
        assert (
            list(JSONModel.objects.filter(attrs__length=2)) ==
            [self.objs[3]]
        )

    def test_length_lookup_chains(self):
        assert (
            list(JSONModel.objects.filter(attrs__length__range=[0, 10])) ==
            self.objs
        )

    def test_shallow_list_lookup(self):
        assert (
            list(JSONModel.objects.filter(attrs__0=1)) ==
            [self.objs[3]]
        )

    def test_shallow_obj_lookup(self):
        assert (
            list(JSONModel.objects.filter(attrs__a='b')) ==
            [self.objs[1], self.objs[2]]
        )

    def test_deep_lookup_objs(self):
        assert (
            list(JSONModel.objects.filter(attrs__k__l='m')) ==
            [self.objs[2]]
        )

    def test_shallow_lookup_obj_target(self):
        assert (
            list(JSONModel.objects.filter(attrs__k={'l': 'm'})) ==
            [self.objs[2]]
        )

    def test_deep_lookup_array(self):
        assert (
            list(JSONModel.objects.filter(attrs__1__0=2)) ==
            [self.objs[3]]
        )

    def test_deep_lookup_mixed(self):
        assert (
            list(JSONModel.objects.filter(attrs__d__1__f='g')) ==
            [self.objs[2]]
        )

    def test_deep_lookup_gt(self):
        assert (
            list(JSONModel.objects.filter(attrs__c__gt=1)) ==
            []
        )
        assert (
            list(JSONModel.objects.filter(attrs__c__lt=5)) ==
            [self.objs[1], self.objs[2]]
        )

    def test_usage_in_subquery(self):
        assert (
            list(JSONModel.objects.filter(
                id__in=JSONModel.objects.filter(attrs__c=1)
            )) ==
            [self.objs[1], self.objs[2]]
        )


class TestCheck(JSONFieldTestCase):

    def test_mutable_default_list(self):
        class InvalidJSONModel(TemporaryModel):
            field = JSONField(default=[])

        errors = InvalidJSONModel.check(actually_check=True)
        assert len(errors) == 1
        assert errors[0].id == 'django_mysql.E017'
        assert 'Do not use mutable defaults for JSONField' in errors[0].msg

    def test_mutable_default_dict(self):
        class InvalidJSONModel(TemporaryModel):
            field = JSONField(default=[])

        errors = InvalidJSONModel.check(actually_check=True)
        assert len(errors) == 1
        assert errors[0].id == 'django_mysql.E017'
        assert 'Do not use mutable defaults for JSONField' in errors[0].msg

    wrapper_path = 'django.db.backends.mysql.base.DatabaseWrapper'

    @mock.patch(wrapper_path + '.is_mariadb', new=True)
    def test_db_not_mysql(self):
        # Uncache cached_property
        for db in connections:
            if 'is_mariadb' in connections[db].__dict__:
                del connections[db].__dict__['is_mariadb']

        class InvalidJSONModel(TemporaryModel):
            field = JSONField()

        errors = InvalidJSONModel.check(actually_check=True)
        assert len(errors) == 1
        assert errors[0].id == 'django_mysql.E016'
        assert "MySQL 5.7+ is required" in errors[0].msg

    @mock.patch(wrapper_path + '.mysql_version', new=(5, 5, 3))
    def test_mysql_old_version(self):
        # Uncache cached_property
        for db in connections:
            if 'mysql_version' in connections[db].__dict__:
                del connections[db].__dict__['mysql_version']

        class InvalidJSONModel(TemporaryModel):
            field = JSONField()

        errors = InvalidJSONModel.check(actually_check=True)
        assert len(errors) == 1
        assert errors[0].id == 'django_mysql.E016'
        assert "MySQL 5.7+ is required" in errors[0].msg


class TestFormField(JSONFieldTestCase):

    def test_model_field_formfield(self):
        model_field = JSONField()
        form_field = model_field.formfield()
        assert isinstance(form_field, forms.JSONField)



# -*- coding:utf-8 -*-
import pytest
from django.db.models import F
from django.test import TestCase

from django_mysql.models import BitAnd, BitOr, BitXor, GroupConcat
from django_mysql.test.utils import override_mysql_variables
from testapp.models import Alphabet, Author


class BitAndTests(TestCase):

    def test_implicit_name(self):
        Alphabet.objects.bulk_create([Alphabet(a=29), Alphabet(a=15)])
        out = Alphabet.objects.all().aggregate(BitAnd('a'))
        assert out == {'a__bitand': 13}

    def test_explicit_name(self):
        Alphabet.objects.bulk_create([Alphabet(a=11), Alphabet(a=24)])
        out = Alphabet.objects.all().aggregate(aaa=BitAnd('a'))
        assert out == {'aaa': 8}

    def test_no_rows(self):
        out = Alphabet.objects.all().aggregate(BitAnd('a'))
        # Manual:
        # "This function returns 18446744073709551615 if there were no matching
        # rows. (This is the value of an unsigned BIGINT value with all bits
        # set to 1.)"
        assert out == {'a__bitand': 18446744073709551615}


class BitOrTests(TestCase):

    def test_implicit_name(self):
        Alphabet.objects.bulk_create([Alphabet(a=1), Alphabet(a=84)])
        out = Alphabet.objects.all().aggregate(BitOr('a'))
        assert out == {'a__bitor': 85}

    def test_explicit_name(self):
        Alphabet.objects.bulk_create([Alphabet(a=29), Alphabet(a=15)])
        out = Alphabet.objects.all().aggregate(aaa=BitOr('a'))
        assert out == {'aaa': 31}

    def test_no_rows(self):
        out = Alphabet.objects.all().aggregate(BitOr('a'))
        # Manual:
        # "This function returns 0 if there were no matching rows."
        assert out == {'a__bitor': 0}


class BitXorTests(TestCase):

    def test_implicit_name(self):
        Alphabet.objects.bulk_create([Alphabet(a=11), Alphabet(a=3)])
        out = Alphabet.objects.all().aggregate(BitXor('a'))
        assert out == {'a__bitxor': 8}

    def test_explicit_name(self):
        Alphabet.objects.bulk_create([Alphabet(a=123), Alphabet(a=456)])
        out = Alphabet.objects.all().aggregate(aaa=BitXor('a'))
        assert out == {'aaa': 435}

    def test_no_rows(self):
        out = Alphabet.objects.all().aggregate(BitXor('a'))
        # Manual:
        # "This function returns 0 if there were no matching rows."
        assert out == {'a__bitxor': 0}


class GroupConcatTests(TestCase):

    def setUp(self):
        self.shakes = Author.objects.create(name='William Shakespeare')
        self.jk = Author.objects.create(name='JK Rowling', tutor=self.shakes)
        self.grisham = Author.objects.create(name='Grisham', tutor=self.shakes)

        self.str_tutee_ids = [str(self.jk.id), str(self.grisham.id)]

    def test_basic_aggregate_ids(self):
        out = self.shakes.tutees.aggregate(tids=GroupConcat('id'))
        concatted_ids = ",".join(self.str_tutee_ids)
        assert out == {'tids': concatted_ids}

    def test_basic_annotate_ids(self):
        concat = GroupConcat('tutees__id')
        shakey = Author.objects.annotate(tids=concat).get(id=self.shakes.id)
        concatted_ids = ",".join(self.str_tutee_ids)
        assert shakey.tids, concatted_ids

    def test_separator(self):
        concat = GroupConcat('id', separator=':')
        out = self.shakes.tutees.aggregate(tids=concat)
        concatted_ids = ":".join(self.str_tutee_ids)
        assert out == {'tids': concatted_ids}

    def test_separator_empty(self):
        concat = GroupConcat('id', separator='')
        out = self.shakes.tutees.aggregate(tids=concat)
        concatted_ids = "".join(self.str_tutee_ids)
        assert out == {'tids': concatted_ids}

    def test_separator_big(self):
        concat = GroupConcat('id', separator='BIG')
        out = self.shakes.tutees.aggregate(tids=concat)
        concatted_ids = "BIG".join(self.str_tutee_ids)
        assert out == {'tids': concatted_ids}

    def test_expression(self):
        concat = GroupConcat(F('id') + 1)
        out = self.shakes.tutees.aggregate(tids=concat)
        concatted_ids = ",".join([
            str(self.jk.id + 1),
            str(self.grisham.id + 1),
        ])
        assert out == {'tids': concatted_ids}

    def test_application_order(self):
        out = (
            Author.objects
                  .exclude(id=self.shakes.id)
                  .aggregate(tids=GroupConcat('tutor_id', distinct=True))
        )
        assert out == {'tids': str(self.shakes.id)}

    @override_mysql_variables(SQL_MODE="ANSI")
    def test_separator_ansi_mode(self):
        concat = GroupConcat('id', separator='>>')
        out = self.shakes.tutees.aggregate(tids=concat)
        concatted_ids = ">>".join(self.str_tutee_ids)
        assert out == {'tids': concatted_ids}

    def test_ordering_invalid(self):
        with pytest.raises(ValueError) as excinfo:
            self.shakes.tutees.aggregate(
                tids=GroupConcat('id', ordering='asceding')
            )
        assert "'ordering' must be one of" in str(excinfo.value)

    def test_ordering_asc(self):
        out = self.shakes.tutees.aggregate(
            tids=GroupConcat('id', ordering='asc')
        )
        assert out == {'tids': ",".join(self.str_tutee_ids)}

    def test_ordering_desc(self):
        out = self.shakes.tutees.aggregate(
            tids=GroupConcat('id', ordering='desc')
        )
        assert out == {'tids': ",".join(reversed(self.str_tutee_ids))}



# -*- coding:utf-8 -*-
import pytest
from django.test import TestCase

from django_mysql.exceptions import TimeoutError
from django_mysql.status import (
    GlobalStatus, SessionStatus, global_status, session_status
)


class BaseStatusTests(TestCase):

    def test_cast_int(self):
        val = global_status._cast('1')
        assert val == 1 and isinstance(val, int)

    def test_cast_float(self):
        val = global_status._cast('1.0')
        assert val == 1.0 and isinstance(val, float)

    def test_cast_on(self):
        val = global_status._cast('ON')
        assert val and isinstance(val, bool)

    def test_cast_off(self):
        val = global_status._cast('OFF')
        assert not val and isinstance(val, bool)

    def test_cast_string(self):
        val = global_status._cast('something')
        assert val == 'something'


class GlobalStatusTests(TestCase):

    def test_get(self):
        running = global_status.get('Threads_running')
        assert running >= 1 and isinstance(running, int)

    def test_get_bad_name(self):
        with pytest.raises(ValueError) as excinfo:
            global_status.get('foo%')
        assert 'wildcards' in str(excinfo.value)

    def test_get_non_existent(self):
        with pytest.raises(KeyError):
            global_status.get('Does_not_exist')

    def test_get_many_empty(self):
        myvars = global_status.get_many([])
        assert myvars == {}

    def test_get_many_existent(self):
        myvars = global_status.get_many(['Threads_running', 'Uptime'])
        assert isinstance(myvars, dict)
        assert 'Threads_running' in myvars
        assert isinstance(myvars['Threads_running'], int)
        assert 'Uptime' in myvars
        assert isinstance(myvars['Uptime'], int)

    def test_get_many_bad_name(self):
        with pytest.raises(ValueError) as excinfo:
            global_status.get_many(['foo%'])
        assert 'wildcards' in str(excinfo.value)

    def test_as_dict(self):
        status_dict = global_status.as_dict()

        assert 'Aborted_clients' in status_dict  # Global-only variable

        assert isinstance(status_dict['Threads_running'], int)
        assert status_dict['Threads_running'] >= 1

    def test_as_dict_prefix(self):
        status_dict = global_status.as_dict()

        status_dict_threads = global_status.as_dict('Threads_')
        assert len(status_dict_threads) < len(status_dict)
        for key in status_dict_threads:
            assert key.startswith('Threads_')

        status_dict_foo = global_status.as_dict('Foo_Non_Existent')
        assert len(status_dict_foo) == 0

    def test_wait_until_load_low(self):
        # Assume tests are running on a non-busy server
        global_status.wait_until_load_low()
        global_status.wait_until_load_low({'Threads_running': 50,
                                           'Threads_connected': 100})

        with pytest.raises(TimeoutError) as excinfo:
            global_status.wait_until_load_low(
                {'Threads_running': -1},  # obviously impossible
                timeout=0.001,
                sleep=0.0005
            )
        message = str(excinfo.value)
        assert 'Threads_running' in message
        assert '-1' in message

        with pytest.raises(TimeoutError) as excinfo:
            global_status.wait_until_load_low(
                {'Threads_running': 1000000,
                 'Uptime': -1},  # obviously impossible
                timeout=0.001,
                sleep=0.0005
            )
        message = str(excinfo.value)
        assert 'Uptime' in message
        assert '-1' in message
        assert 'Threads_running' not in message
        assert '1000000' not in message

    def test_other_databases(self):
        status = GlobalStatus(using='other')

        running = status.get('Threads_running')
        assert isinstance(running, int)
        assert running >= 1


class SessionStatusTests(TestCase):

    def test_get_bytes_received(self):
        bytes_received = session_status.get('Bytes_received')
        assert bytes_received >= 0 and isinstance(bytes_received, int)

        bytes_received_2 = session_status.get('Bytes_received')
        assert bytes_received_2 >= bytes_received

    def test_get_last_query_cost(self):
        cost = session_status.get('Last_query_cost')
        assert cost >= 0.0 and isinstance(cost, float)

    def test_get_bad_name(self):
        with pytest.raises(ValueError) as excinfo:
            session_status.get('foo%')
        assert 'wildcards' in str(excinfo.value)

    def test_get_non_existent(self):
        with pytest.raises(KeyError):
            session_status.get('Does_not_exist')

    def test_as_dict(self):
        status_dict = session_status.as_dict()

        assert 'Compression' in status_dict  # status only variable (on 5.7+)

        assert isinstance(status_dict['Threads_running'], int)
        assert status_dict['Threads_running'] >= 1

    def test_other_databases(self):
        status = SessionStatus(using='other')
        running = status.get('Threads_running')
        assert isinstance(running, int)
        assert running >= 1



# -*- coding:utf-8 -*-
from __future__ import unicode_literals

import functools

from django.conf import settings
from django.db.backends.mysql.base import CursorWrapper
from django.utils.functional import cached_property

from django_mysql.rewrite_query import REWRITE_MARKER, rewrite_query
from django_mysql.utils import connection_is_mariadb


def patch():
    # Fine to patch straight on since it's a cached_property descriptor
    patch_ConnectionWrapper_is_mariadb()

    # Depends on setting
    if getattr(settings, 'DJANGO_MYSQL_REWRITE_QUERIES', False):
        patch_CursorWrapper_execute()


def patch_ConnectionWrapper_is_mariadb():
    from django.db.backends.mysql.base import DatabaseWrapper

    @cached_property
    def is_mariadb(self):
        return connection_is_mariadb(self)

    DatabaseWrapper.is_mariadb = is_mariadb


def patch_CursorWrapper_execute():

    # Be idemptotent
    if getattr(CursorWrapper, '_has_django_mysql_execute', False):
        return

    orig_execute = CursorWrapper.execute

    @functools.wraps(orig_execute)
    def execute(self, sql, args=None):
        if (
            getattr(settings, 'DJANGO_MYSQL_REWRITE_QUERIES', False) and
            REWRITE_MARKER in sql
        ):
            sql = rewrite_query(sql)
        return orig_execute(self, sql, args)

    CursorWrapper.execute = execute
    CursorWrapper._has_django_mysql_execute = True



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations

from django_mysql.models import EnumField


class Migration(migrations.Migration):

    dependencies = [
        ('testapp', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='EnumDefaultModel',
            name='field',
            field=EnumField(choices=[
                ('lion', 'Lion'),
                ('tiger', 'Tiger'),
                ('bear', 'Bear'),
                'oh my!'
            ])
        ),
    ]
